	
// MPIJob.h : Declaration of the CMPIJob

#ifndef __MPIJOB_H_
#define __MPIJOB_H_

#include "resource.h"       // main symbols
#include "..\Common\MPIJobLog.h"

// Forward declarations
struct ExtendJobThreadArg;
struct OutputNode;
struct InputNode;

/////////////////////////////////////////////////////////////////////////////
// CMPIJob
class ATL_NO_VTABLE CMPIJob : 
	public CComObjectRootEx<CComMultiThreadModel>,
	public CComCoClass<CMPIJob, &CLSID_MPIJob>,
	public IDispatchImpl<IMPIJob, &IID_IMPIJob, &LIBID_MPIJOBSERVERLib>
{
public:
	CMPIJob()
	{
		m_pUnkMarshaler = NULL;
		m_hThreads = NULL;
		m_pThreadArgs = NULL;
		m_nIproc = 0;
		m_nNproc = 0;
		m_nNumThreads = 0;
		m_pszJobID[0] = L'\0';
		m_pszUser[0] = L'\0';
		m_pszExe[0] = L'\0';
		m_pszNic[0] = L'\0';
		m_pszWorkingDir[0] = L'.';
		m_pszWorkingDir[1] = L'\0';
		m_pszEnv[0] = L'\0';
		m_hTempFile = NULL;
		m_hOutputMutex = CreateMutex(NULL, FALSE, NULL);
		m_hOutputEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
		m_hInputMutex = CreateMutex(NULL, FALSE, NULL);
		m_hInputEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
		m_hAbortEvent = CreateEvent(NULL, TRUE, TRUE, NULL);
		m_hBreakEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
		m_pOutList = NULL;
		m_pOutListTail = NULL;
		m_pInList = NULL;
		m_pInListTail = NULL;
		m_nOutputStreamsRemaining = 1;
		m_bInconsistentState = false;
		m_bLaunchThreadStarted = false;
		m_bCreateJobCalled = false;
		m_nLaunchFlags = 0;
		m_bNoMPI = false;
		m_bFixedPort = false;
		m_nSMPProcs = 1;
		DWORD len = 100;
		if (!GetComputerNameW(m_pszLocalHost, &len))
			m_pszLocalHost[0] = L'\0';
	}
	~CMPIJob();

DECLARE_REGISTRY_RESOURCEID(IDR_MPIJOB)
DECLARE_GET_CONTROLLING_UNKNOWN()

DECLARE_PROTECT_FINAL_CONSTRUCT()

BEGIN_COM_MAP(CMPIJob)
	COM_INTERFACE_ENTRY(IMPIJob)
	COM_INTERFACE_ENTRY(IDispatch)
	COM_INTERFACE_ENTRY_AGGREGATE(IID_IMarshal, m_pUnkMarshaler.p)
END_COM_MAP()

	HRESULT FinalConstruct()
	{
		return CoCreateFreeThreadedMarshaler(
			GetControllingUnknown(), &m_pUnkMarshaler.p);
	}

	void FinalRelease()
	{
		m_pUnkMarshaler.Release();
	}

	CComPtr<IUnknown> m_pUnkMarshaler;

// IMPIJob
public:
	STDMETHOD(SendBreak)(long *error, BSTR *berror_msg);
	STDMETHOD(SetEnvironmentVariables)(BSTR bEnv, long *error, BSTR *berror_msg);
	STDMETHOD(PutJobInput)(VARIANT vInput, long to, long *error, BSTR *berror_msg);
	STDMETHOD(SetWorkingDirectory)(BSTR bDirectory, long *error, BSTR *berror_msg);
	STDMETHOD(Abort)(long *error, BSTR *berror_msg);
	STDMETHOD(GetJobOutput)(VARIANT *vOutput, long *from, long *more, long *error, BSTR *berror_msg);
	STDMETHOD(ExtendLaunchJob)(BSTR bCommonArgs, BSTR bAccount, BSTR bPassword, BSTR bRootHost, BSTR bRootPort, long flags, long *error, BSTR *berror_msg);
	STDMETHOD(LaunchJob)(BSTR bCommonArgs, BSTR bAccount, BSTR bPassword, long flags, long *error, BSTR *berror_msg);
	STDMETHOD(FinishExeTransfer)(long *error, BSTR *berror_msg);
	STDMETHOD(TransferExeData)(VARIANT vData, long *error, BSTR *berror_msg);
	STDMETHOD(StartExeTransfer)(long *error, BSTR *berror_msg);
	STDMETHOD(SetExecutable)(BSTR bExecutable, long *error, BSTR *berror_msg);
	STDMETHOD(DestroyJob)(long *error, BSTR *berror_msg);
	STDMETHOD(ExtendJob)(long iproc, long nproc, BSTR bJobID, VARIANT vHosts, long nSMPProcs, BSTR bComNic, long *error, BSTR *berror_msg);
	STDMETHOD(CreateJob)(VARIANT vHosts, VARIANT vSMPInfo, long nSMPProcs, BSTR bComNic, long *error, BSTR *berror_msg);

	WCHAR m_pszJobID[100], m_pszUser[100], m_pszExe[MAX_PATH], m_pszNic[100], m_pszWorkingDir[MAX_PATH], m_pszEnv[1024], m_pszLocalHost[100];
	HANDLE *m_hThreads;
	ExtendJobThreadArg *m_pThreadArgs;
	long m_nIproc, m_nNumThreads, m_nNproc;
	HANDLE m_hTempFile;

	OutputNode *m_pOutList, *m_pOutListTail;
	HANDLE m_hOutputEvent, m_hOutputMutex;
	InputNode *m_pInList, *m_pInListTail;
	HANDLE m_hInputEvent, m_hInputMutex;
	int m_nOutputStreamsRemaining;
	HANDLE m_hAbortEvent, m_hBreakEvent;
	long m_nLaunchFlags, m_nSMPProcs;

	bool m_bInconsistentState, m_bLaunchThreadStarted, m_bCreateJobCalled, m_bNoMPI, m_bFixedPort;
};

struct ExtendJobThreadArg
{
	VARIANT *vHosts;
	WCHAR pszHost[100];
	long nIproc, nNproc, nSMPProcs;
	HANDLE hCommandEvent, hCommandCompletedEvent;
	int nCommand;
	void *pCommandData;
	HRESULT nResult;
	WCHAR wErrorMsg[256];
	CMPIJob *pJob;
};

struct OutputNode
{
	OutputNode(): data(NULL), size(0), from(0), next(NULL) {};

	unsigned char *data;
	unsigned long size;
	int from;
	OutputNode *next;
};

struct InputNode
{
	unsigned char *data;
	unsigned long size;
	InputNode *next;
};

struct AbortThreadArg
{
	HANDLE hAbortEvent;
	bool bNormalExit;
	IStream **ppStream;
};

struct PutInputThreadArg
{
	IStream **ppStream;
	CMPIJob *pJob;
};

struct CollectOutputThreadArg
{
	long nIproc;
	CMPIJob *pJob;
	IStream **ppStream;
};

struct LaunchJobThreadArg
{
	BSTR bJobID;
	BSTR bCmdLine;
	BSTR bAccount;
	BSTR bPassword;
	BSTR bExtra;
	BSTR bNic;
	long nIproc, nNproc;
	BSTR bRootHost;
	long nRootPort;
	CMPIJob *pJob;

	long pid;
	HANDLE hReadyEvent;
	HRESULT hr;
	WCHAR wErrorMsg[256];
	long shmLow, shmHigh;
};

#endif //__MPIJOB_H_
