// MPISystem.cpp : Implementation of CMPISystem
#include "stdafx.h"
#include "MPISystemServer.h"
#include "MPISystem.h"
#include "..\Common\DCOMPerm\DCOMPerm.h"
#include "..\Common\MPIJobDefs.h"
#include "..\Common\MPIJobLog.h"

/////////////////////////////////////////////////////////////////////////////
// CMPISystem


STDMETHODIMP CMPISystem::ConfigAccess(BSTR bIIDString, BSTR bPrincipal, long flags, long *error, BSTR *berror_msg)
{
	TCHAR tIIDString[100], tPrincipal[100];
	bool bAccess, bSet, bPermit;

	try{
#ifdef UNICODE
	wcscpy(tIIDString, bIIDString);
	wcscpy(tPrincipal, bPrincipal);
#else
	wcstombs(tIIDString, bIIDString, wcslen(bIIDString)+1);
	wcstombs(tPrincipal, bPrincipal, wcslen(bPrincipal)+1);
#endif

	if ((flags & MPICONFIGPERMIT) && (flags & MPICONFIGDENY))
	{
		*error = 1;
		SysReAllocString(berror_msg, L"Invalid argument: Permit and Deny cannot be simultaneously set.");
		return S_OK;
	}

	if ((flags & MPICONFIGSET) && (flags & MPICONFIGREMOVE))
	{
		*error = 1;
		SysReAllocString(berror_msg, L"Invalid argument: Set and Remove cannot be simultaneously set.");
		return S_OK;
	}

	if (flags & MPICONFIGPERMIT)
		bPermit = true;
	if (flags & MPICONFIGDENY)
		bPermit = false;
	if (flags & MPICONFIGSET)
		bSet = true;
	if (flags & MPICONFIGREMOVE)
		bSet = false;

	if (flags & MPICONFIGACCESS)
	{
		bAccess = true;
		DCOMPermissions(tIIDString, bAccess, tPrincipal, bSet, bPermit);
	}

	if (flags & MPICONFIGLAUNCH)
	{
		bAccess = false;
		DCOMPermissions(tIIDString, bAccess, tPrincipal, bSet, bPermit);
	}

	}catch(...)
	{
		*error = 1;
		SysReAllocString(berror_msg, L"Exception thrown in ConfigAccess.\n");
		LogMsg(TEXT("Exception thrown in ConfigAccess.\n"));
	}
	return S_OK;
}

STDMETHODIMP CMPISystem::ConfigRunAs(BSTR bIIDString, BSTR bPrincipal, BSTR bPassword, long *error, BSTR *berror_msg)
{
	TCHAR tIIDString[100], tPrincipal[100], tPassword[100];

	try{
#ifdef UNICODE
	wcscpy(tIIDString, bIIDString);
	wcscpy(tPrincipal, bPrincipal);
	wcscpy(tPassword, bPassword);
#else
	wcstombs(tIIDString, bIIDString, wcslen(bIIDString)+1);
	wcstombs(tPrincipal, bPrincipal, wcslen(bPrincipal)+1);
	wcstombs(tPassword, bPassword, wcslen(bPassword)+1);
#endif

	DCOMSetRunAs(tIIDString, tPrincipal, tPassword);
	}catch(...)
	{
		*error = 1;
		SysReAllocString(berror_msg, L"Exception thrown in ConfigRunAs.\n");
		LogMsg(TEXT("Exception thrown in ConfigRunAs.\n"));
	}
	return S_OK;
}

STDMETHODIMP CMPISystem::Reboot(long *error, BSTR *berror_msg)
{
	HANDLE token;
	TOKEN_PRIVILEGES token_privileges;
	try{
	if ( OpenProcessToken( 
			GetCurrentProcess(),
			TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY,
			&token ) == 0 ) 
	{
		*error = GetLastError();
		SysReAllocString(berror_msg, L"OpenProcessToken failed.\n");
		return S_OK;
	}
	
	if ( LookupPrivilegeValue(
			0, SE_SHUTDOWN_NAME,
			&token_privileges.Privileges[0].Luid ) == 0 ) 
	{
		*error = GetLastError();
		SysReAllocString(berror_msg, L"LookupPrivilegeValue failed.\n");
		return S_OK;
	}
	
	token_privileges.PrivilegeCount = 1;
	token_privileges.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;
	if ( AdjustTokenPrivileges(token, FALSE,
		&token_privileges, 0, 0, 0) == 0 ) 
	{
		*error = GetLastError();
		SysReAllocString(berror_msg, L"AdjustTokenPrivileges failed.\n");
		return S_OK;
	}
	
	if ( InitiateSystemShutdown(NULL, NULL, 0, TRUE, TRUE) == 0 )
	{
		*error = GetLastError();
		SysReAllocString(berror_msg, L"InitiateSystemShutdown failed.\n");
		return S_OK;
	}
	}catch(...)
	{
		*error = 1;
		SysReAllocString(berror_msg, L"Exception thrown in Reboot.\n");
		LogMsg(TEXT("Exception thrown in Reboot.\n"));
	}
	return S_OK;
}

STDMETHODIMP CMPISystem::ClearAndBurnAllJobs(long *error, BSTR *berror_msg)
{
	return S_OK;
}
