#include "nt_global_cpp.h"
#include <stdio.h>

HANDLE g_hCommThread;
HANDLE g_hCommEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
int g_nCommCommand;

SOCKET g_socket_to_add;
HANDLE g_socket_to_add_event;
int    g_socket_to_add_rank;
HANDLE g_hSocketAddedEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
HANDLE g_hAddSocketMutex = CreateMutex(NULL, FALSE, NULL);

#define NT_MPI_MAX_NUM_WAIT_THREADS MAXIMUM_WAIT_OBJECTS

struct WaitThreadArg
{
	HANDLE hStopEvent, hStartEvent, hSignal, *pEvent;
	int num_events;
};

HANDLE g_hWaitThreads[NT_MPI_MAX_NUM_WAIT_THREADS];
HANDLE g_hStopWaitThreadEvents[NT_MPI_MAX_NUM_WAIT_THREADS];
HANDLE g_hStartWaitThreadEvents[NT_MPI_MAX_NUM_WAIT_THREADS];
HANDLE g_hWaitThreadSignalEvents[NT_MPI_MAX_NUM_WAIT_THREADS];
WaitThreadArg g_pWaitThreadArgs[NT_MPI_MAX_NUM_WAIT_THREADS];
int g_nNumWaitHandles = 0;
int g_nNumWaitThreads = 0;
bool g_bEventsChanged = false;

// Function name	: WaitThread
// Description	    : 
// Return type		: void 
// Argument         : WaitThreadArg *arg
void WaitThread(WaitThreadArg *arg)
{
	HANDLE hEvents[MAXIMUM_WAIT_OBJECTS];
	hEvents[0] = arg->hStopEvent;
	memcpy(&hEvents[1], arg->pEvent, arg->num_events*sizeof(HANDLE));
	while (true)
	{
		if (arg->num_events == 0)
			return;
		if (g_bEventsChanged)
			memcpy(&hEvents[1], arg->pEvent, arg->num_events*sizeof(HANDLE));
		WaitForMultipleObjects(arg->num_events+1, hEvents, FALSE, INFINITE);
		SetEvent(arg->hSignal);
		WaitForSingleObject(arg->hStartEvent, INFINITE);
		ResetEvent(arg->hStartEvent);
	}
}

// Function name	: MyWaitForMultipleObjects
// Description	    : 
// Return type		: DWORD 
// Argument         : int num_handles
// Argument         : HANDLE *ahEvent
DWORD MyWaitForMultipleObjects(int num_handles, HANDLE *ahEvent)
{
	// TODO set g_bEventsChanged when there is a socket change
	int i;
	if (num_handles < MAXIMUM_WAIT_OBJECTS)
		return WaitForMultipleObjects(num_handles, ahEvent, FALSE, INFINITE);
	else
	{
		if (g_bEventsChanged)
		{
			int num_left = num_handles;
			int cur_thread = 0;
			HANDLE *pEvent = ahEvent;
			while (num_left)
			{
				if (cur_thread <= g_nNumWaitThreads)
				{
					g_pWaitThreadArgs[cur_thread].pEvent = pEvent;
					g_pWaitThreadArgs[cur_thread].hStopEvent = g_hStopWaitThreadEvents[cur_thread];
					g_pWaitThreadArgs[cur_thread].hSignal = g_hWaitThreadSignalEvents[cur_thread];
					g_pWaitThreadArgs[cur_thread].hStartEvent = g_hStartWaitThreadEvents[cur_thread];
					if (num_left > MAXIMUM_WAIT_OBJECTS-1)
					{
						g_pWaitThreadArgs[cur_thread].num_events = MAXIMUM_WAIT_OBJECTS - 1;
						num_left = num_left + 1 - MAXIMUM_WAIT_OBJECTS;
					}
					else
					{
						g_pWaitThreadArgs[cur_thread].num_events = num_left;
						num_left = 0;
					}
					ResetEvent(g_hStopWaitThreadEvents[cur_thread]);
				}
				else
				{
					g_pWaitThreadArgs[cur_thread].pEvent = pEvent;
					g_pWaitThreadArgs[cur_thread].hStopEvent = g_hStopWaitThreadEvents[cur_thread] = CreateEvent(NULL, TRUE, FALSE, NULL);
					g_pWaitThreadArgs[cur_thread].hSignal = g_hWaitThreadSignalEvents[cur_thread] = CreateEvent(NULL, TRUE, FALSE, NULL);
					g_pWaitThreadArgs[cur_thread].hStartEvent = g_hStartWaitThreadEvents[cur_thread] = CreateEvent(NULL, TRUE, FALSE, NULL);
					if (num_left > MAXIMUM_WAIT_OBJECTS-1)
					{
						g_pWaitThreadArgs[cur_thread].num_events = MAXIMUM_WAIT_OBJECTS - 1;
						num_left = num_left + 1 - MAXIMUM_WAIT_OBJECTS;
					}
					else
					{
						g_pWaitThreadArgs[cur_thread].num_events = num_left;
						num_left = 0;
					}
					DWORD dwThreadID;
					g_hWaitThreads[cur_thread] = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)WaitThread, &g_pWaitThreadArgs[cur_thread], 0, &dwThreadID);
					g_nNumWaitThreads++;
					if (g_hWaitThreads[cur_thread] == NULL)
						nt_error("CreateThread failed in MyWaitForMultipleObjects", 1);
				}
				cur_thread++;
			}
		}
		
		for (i=0; i<g_nNumWaitThreads; i++)
			SetEvent(g_hStartWaitThreadEvents[i]);
		WaitForMultipleObjects(g_nNumWaitThreads, g_hWaitThreadSignalEvents, FALSE, INFINITE);
		for (i=0; i<g_nNumWaitThreads; i++)
			SetEvent(g_hStopWaitThreadEvents[i]);
		WaitForMultipleObjects(g_nNumWaitThreads, g_hWaitThreadSignalEvents, TRUE, INFINITE);
		for (i=0; i<g_nNumWaitThreads; i++)
			ResetEvent(g_hWaitThreadSignalEvents[i]);
		g_bEventsChanged = false;
	}
	return WAIT_OBJECT_0;
}

// Function name	: CommunicationThread
// Description	    : 
// Return type		: void 
// Argument         : HANDLE hReadyEvent
void CommunicationThread(HANDLE hReadyEvent)
{
	SOCKET listen_socket;
	HANDLE *ahEvent;				// array of events to wait on
	int *anMap;						// array of integers mapping from the event to the socket it corresponds to
	int error = 0, num_handles=2;
	SOCKET temp_socket;
	WSAEVENT temp_event;
	DWORD ret_val;
	int remote_iproc;
	int tag, length, gindex;
	void *buffer;
	int i;
	BOOL opt;
	char add_socket_ack;
	MessageQueue::MsgQueueElement *pElement;

	// Allocate space for nproc connections + listen thread + command event
	anMap = new int[g_nNproc+2];
	anMap[0] = -1;
	anMap[1] = -1;
	ahEvent = new HANDLE[g_nNproc+2];
	ahEvent[0] = g_hCommEvent;

	// create a listening socket
	error = NT_Tcp_shm_create_bind_socket(&listen_socket, &ahEvent[1]);
	if (error)
		nt_error("CommunicationThread: NT_Tcp_shm_create_bind_socket failed", 1);

	// associate listen_socket_event with listen_socket
	if (WSAEventSelect(listen_socket, ahEvent[1], FD_ACCEPT) == SOCKET_ERROR)
		nt_error("CommunicationThread: WSAEventSelect failed for listen_socket", 1);

	if (listen(listen_socket, SOMAXCONN) == SOCKET_ERROR)
		nt_error("CommunicationThread: listen failed", WSAGetLastError());

	// get the port and local hostname for the listening socket
	error = NT_Tcp_shm_get_sock_info(listen_socket, g_pProcTable[g_nIproc].host, &g_pProcTable[g_nIproc].listen_port);
	if (error)
		nt_error("CommunicationThread: Unable to get host and port of listening socket", error);

	// Signal that the port number is valid
	SetEvent(hReadyEvent);

	g_bEventsChanged = true;
	try{
	while (true)
	{
		ret_val = MyWaitForMultipleObjects(num_handles, ahEvent);
		if (ret_val == WAIT_FAILED)
		{
			nt_error("CommunicationThread: Wait failed", GetLastError());
			return;
		}

		// Event[0] is the event used by other threads in this process to communicate with this thread
		if (WaitForSingleObject(ahEvent[0], 0) == WAIT_OBJECT_0)
		{
			switch (g_nCommCommand)
			{
			case NT_COMM_CMD_ADD_SOCKET:
				DPRINTF(("process %d: Add socket command.\n", g_nIproc));
				if((g_socket_to_add_rank >= 0) && (g_socket_to_add_rank < g_nNproc))
				{
					g_pProcTable[g_socket_to_add_rank].sock = g_socket_to_add;
					g_pProcTable[g_socket_to_add_rank].sock_event = g_socket_to_add_event;
					ahEvent[num_handles] = g_socket_to_add_event;
					anMap[num_handles] = g_socket_to_add_rank;
					num_handles++;
					g_bEventsChanged = true;
					
					ResetEvent(g_hCommEvent);
					SetEvent(g_hSocketAddedEvent);
				}
				else
					MakeErrMsg(1, "g_socket_to_add_rank out of range: %d", g_socket_to_add_rank);
				break;
			case NT_COMM_CMD_EXIT:
				DPRINTF(("process %d: Exit command.\n", g_nIproc));
				CloseHandle(g_hAddSocketMutex);
				// close g_hCommEvent
				CloseHandle(ahEvent[0]); 
				// close the listening socket
				closesocket(listen_socket);
				WSACloseEvent(ahEvent[1]);
				delete anMap;
				delete ahEvent;
				ExitThread(0);
				break;
			}
		}

		// Event[1] is the listen socket event, which is signalled when other processes whish to establish a socket connection with this process
		if (WaitForSingleObject(ahEvent[1], 0) == WAIT_OBJECT_0)
		{
			///DPRINTF(("process %d: listen_socket signalled.\n", g_nIproc));
			i=0;
			opt = TRUE;
			// Something in my code is causing the listen_socket event to fail to be reset by the accept call.
			// For now I manually reset it here.
			WSAResetEvent(ahEvent[1]);
			temp_socket = accept(listen_socket, NULL, NULL);
			if (temp_socket != INVALID_SOCKET)
			{
				// Create an event and associate it with the newly accepted socket
				if (setsockopt(temp_socket, IPPROTO_TCP, TCP_NODELAY, (char*)&opt, sizeof(BOOL)) == SOCKET_ERROR)
					nt_error("setsockopt failed", WSAGetLastError());
				if ((temp_event = WSACreateEvent()) == WSA_INVALID_EVENT)
					nt_error("WSACreateEvent failed after accepting socket", WSAGetLastError());
				if (WSAEventSelect(temp_socket, temp_event, FD_READ | FD_CLOSE) == SOCKET_ERROR)
					nt_error("WSAEventSelect failed after accepting socket", WSAGetLastError());
				
				// Receive the rank of the remote process
				if (ret_val = ReceiveBlocking(temp_socket, temp_event, (char*)&remote_iproc, sizeof(int), 0))
					nt_error("ReceiveBlocking remote_iproc failed after accepting socket", ret_val);
				
				if (remote_iproc >= 0 && remote_iproc < g_nNproc)
				{
					if (WaitForSingleObject(g_hAddSocketMutex, 0) == WAIT_OBJECT_0)
					{
						if (g_pProcTable[remote_iproc].sock == INVALID_SOCKET)
						{
							add_socket_ack = 1;
							if (SendBlocking(temp_socket, &add_socket_ack, 1, 0) == SOCKET_ERROR)
								MakeErrMsg(WSAGetLastError(), "send add_socket_ack(1) failed for socket %d", remote_iproc);
							// Insert the information in g_pProcTable
							g_pProcTable[remote_iproc].sock_event = temp_event;
							g_pProcTable[remote_iproc].sock = temp_socket;
							
							// Insert the information in the local arrays
							ahEvent[num_handles] = temp_event;
							anMap[num_handles] = remote_iproc;
							ResetEvent(ahEvent[num_handles]);
							if (WSAEventSelect(temp_socket, temp_event, FD_READ | FD_CLOSE) == SOCKET_ERROR)
								nt_error("WSAEventSelect failed after accepting socket", WSAGetLastError());
							num_handles++;
							g_bEventsChanged = true;
							
							DPRINTF(("process %d: socket accepted and inserted in location %d\n", g_nIproc, remote_iproc));
						}
						else
						{
							add_socket_ack = 0;
							if (SendBlocking(temp_socket, &add_socket_ack, 1, 0) == SOCKET_ERROR)
								MakeErrMsg(WSAGetLastError(), "send add_socket_ack(0) failed for socket %d", remote_iproc);
							NT_Tcp_shm_closesocket(temp_socket, temp_event);
						}
						ReleaseMutex(g_hAddSocketMutex);
					}
					else
					{
						if (g_nIproc > remote_iproc)
						{
							add_socket_ack = 1;
							if (SendBlocking(temp_socket, &add_socket_ack, 1, 0) == SOCKET_ERROR)
								MakeErrMsg(WSAGetLastError(), "send add_socket_ack(1) failed for socket %d", remote_iproc);
						
							// Insert the information in the local arrays
							ahEvent[num_handles] = temp_event;
							anMap[num_handles] = remote_iproc;
							ResetEvent(ahEvent[num_handles]);
							if (WSAEventSelect(temp_socket, temp_event, FD_READ | FD_CLOSE) == SOCKET_ERROR)
								nt_error("WSAEventSelect failed after accepting socket", WSAGetLastError());
							num_handles++;
							g_bEventsChanged = true;
							
							// Insert the information in g_pProcTable
							g_pProcTable[remote_iproc].sock_event = temp_event;
							g_pProcTable[remote_iproc].sock = temp_socket;
							
							DPRINTF(("process %d: socket accepted and inserted in location %d\n", g_nIproc, remote_iproc));
						}
						else
						{
							add_socket_ack = 0;
							if (SendBlocking(temp_socket, &add_socket_ack, 1, 0) == SOCKET_ERROR)
								MakeErrMsg(1, "send add_socket_ack(0) failed for socket %d", remote_iproc);
							NT_Tcp_shm_closesocket(temp_socket, temp_event);
						}
					}
				}
				else
				{
					MakeErrMsg(1, "CommunicationThread: Process out of range, remote_iproc: %d\n", remote_iproc);
					return;
				}
			}
			else
			{
				error = WSAGetLastError();
				if (error != WSAEWOULDBLOCK)
				{
					nt_error("CommunicationThread: accept failed", error);
					return;
				}
			}
		}

		// Event[2..n] are the events associated with point to point socket connections
		// This linear check of all the sockets may add to the latency cost signifigantly if you have a very low latency tcp interface.
		// I count down from the end so I can remove sockets that are closing.
		for (i=num_handles - 1; i>1; i--)
		{
			if (WaitForSingleObject(ahEvent[i], 0) == WAIT_OBJECT_0)
			{
				DPRINTF(("process %d: communication socket signalled: %d(%d)\n", g_nIproc, anMap[i], i));
				gindex = anMap[i];
				if ((gindex < 0) || (gindex >= g_nNproc))
					MakeErrMsg(1, "Index out of range: anMap[%d] = %d", i, gindex);
				//DPRINTF(("Receiving message from %i\n", gindex));
				if (ret_val = ReceiveBlockingTimeout(g_pProcTable[gindex].sock, g_pProcTable[gindex].sock_event, (char*)&tag, sizeof(int), 0, 300))
				{
					if (ret_val == 1)
					{
						if (g_bInNT_Tcp_shm_End)
						{
							DPRINTF(("process %d: CommunicationThread: ReceiveBlocking failed, closing socket %d\n", g_nIproc, gindex));
							NT_Tcp_shm_closesocket(g_pProcTable[gindex].sock, g_pProcTable[gindex].sock_event);
							g_pProcTable[gindex].sock = INVALID_SOCKET;
							g_pProcTable[gindex].sock_event = NULL;
							num_handles--;
							g_bEventsChanged = true;
							for (int j=i; j<num_handles; j++)
							{
								anMap[j] = anMap[j+1];
								ahEvent[j] = ahEvent[j+1];
							}
							continue;
						}
						MakeErrMsg(1, "Socket %d closed, terminating process %d", gindex, g_nIproc);
					}
					if (ret_val == WSA_WAIT_TIMEOUT)
					{
						// For some reason the socket remains signalled after the first recv call. This leads to entering
						// this piece of code a second time when there is no data available on the socket.  I put a short 
						// timeout on the first blocking receive call to detect this.
						//printf("***Timeout:socket:%d***", anMap[i]);fflush(stdout);
						continue;
					}
					if (ret_val) MakeErrMsg(ret_val, "CommunicationThread: Receiving tag failed from socket %d", gindex);
				}
				DPRINTF(("process %d: communication socket received tag: %d\n", g_nIproc, tag));
				
				if (ret_val = ReceiveBlocking(g_pProcTable[gindex].sock, g_pProcTable[gindex].sock_event, (char*)&length, sizeof(int), 0))
					MakeErrMsg(ret_val, "CommunicationThread: Receiving length failed from socket %d", gindex);
				DPRINTF(("process %d: communication socket received length: %d\n", g_nIproc, length));

				buffer = g_MsgQueue.GetBuffer(tag, length, gindex, &pElement);
				if (buffer == NULL)
					nt_error("CommunicationThread: memory allocation failed.", 1);
				if (ret_val = ReceiveBlocking(g_pProcTable[gindex].sock, g_pProcTable[gindex].sock_event, (char*)buffer, length, 0))
					MakeErrMsg(ret_val, "CommunicationThread: Receiving buffer failed from socket %d", gindex);
				//SetEvent(pElement->hEvent);
				g_MsgQueue.SetElementEvent(pElement);

				ResetEvent(ahEvent[i]);
				WSAEventSelect(g_pProcTable[gindex].sock, ahEvent[i], FD_READ | FD_CLOSE);

				DPRINTF(("process %d: Message(%d, len %d) inserted in queue.\n", g_nIproc, tag, length));
			}
		}
	}
	}catch(...)
	{
		nt_error("Exception caught in CommunicationThread.", 1);
	}
}

// Function name	: ConnectTo
// Description	    : 
// Return type		: int 
// Argument         : int remote_iproc
int ConnectTo(int remote_iproc)
{
	char ack = 0;
	int ret_val;
	BOOL opt = TRUE;
	int i=0;

	if (remote_iproc < 0 || remote_iproc >= g_nNproc)
	{
		MakeErrMsg(1, "ConnectTo failed, invalid remote_iproc: %d\n", remote_iproc);
		return 0;
	}

	if (WaitForSingleObject(g_hAddSocketMutex, 3000) == WAIT_TIMEOUT)
		MakeErrMsg(1, "ConnectTo %d failed, wait for AddSocketMutex timed out", remote_iproc);

	if (g_pProcTable[remote_iproc].sock != INVALID_SOCKET)
	{
		ReleaseMutex(g_hAddSocketMutex);
		return 1;
	}

	GetProcessConnectInfo(remote_iproc);

	HOSTENT *hostEnt;
	unsigned long nic_addr = INADDR_ANY;
	hostEnt = gethostbyname(g_pProcTable[remote_iproc].host);
	if (hostEnt != NULL)
		nic_addr = *((unsigned long*)hostEnt->h_addr_list[0]);
	// Create a socket and connect to process 'remote_iproc'
	// create the event
	g_socket_to_add_event = WSACreateEvent();
	if (g_socket_to_add_event == WSA_INVALID_EVENT)
		nt_error("WSACreateEvent failed in ConnectTo", WSAGetLastError());
	// create the socket
	g_socket_to_add = socket(PF_INET, SOCK_STREAM, 0);
	if (g_socket_to_add == INVALID_SOCKET)
		nt_error("socket failed in ConnectTo", WSAGetLastError());

	DPRINTF(("connecting to %s on %d\n", g_pProcTable[remote_iproc].host, g_pProcTable[remote_iproc].listen_port));
	if (ret_val = NT_Tcp_shm_connect(g_socket_to_add, g_pProcTable[remote_iproc].host, g_pProcTable[remote_iproc].listen_port))
		nt_error("NT_Tcp_shm_connect failed in ConnectTo", ret_val);
	
	if (setsockopt(g_socket_to_add, IPPROTO_TCP, TCP_NODELAY, (char*)&opt, sizeof(BOOL)) == SOCKET_ERROR)
		nt_error("setsockopt failed in ConnectTo", WSAGetLastError());

	if (WSAEventSelect(g_socket_to_add, g_socket_to_add_event, FD_READ | FD_CLOSE) == SOCKET_ERROR)
		nt_error("WSAEventSelect failed in ConnectTo", WSAGetLastError());
	
	// Send my rank so the remote side knows who is connecting
	if (SendBlocking(g_socket_to_add, (char*)&g_nIproc, sizeof(int), 0) == SOCKET_ERROR)
		nt_error("send g_nIproc failed in ConnectTo", WSAGetLastError());

	// Receive an ack determining whether the connection was added to the list or not
	if (ret_val = ReceiveBlocking(g_socket_to_add, g_socket_to_add_event, &ack, 1, 0))
		MakeErrMsg(ret_val, "ConnectTo failed to receive ack for socket %d", remote_iproc);

	if (ack == 1)
	{
		// Signal the CommunicationThread to add the new socket	
		g_socket_to_add_rank = remote_iproc;
		//DPRINTF(("process %d: g_socket_to_add_event: %u g_socket_to_add_rank: %d\n", g_nIproc, g_socket_to_add_event, g_socket_to_add_rank));
		ResetEvent(g_hSocketAddedEvent);
		g_nCommCommand = NT_COMM_CMD_ADD_SOCKET;
		SetEvent(g_hCommEvent);
		if (WaitForSingleObject(g_hSocketAddedEvent, 10000) == WAIT_TIMEOUT)
			MakeErrMsg(1, "wait for g_hSocketAddedEvent timed out for connection %d", remote_iproc);

		DPRINTF(("process %d: established connection to %d\n", g_nIproc, remote_iproc));
	}
	else
	{
		// The listener determined this side to be the loser in a race condition
		// So close the socket and wait for the socket created in another thread
		// to be inserted in the proc table.
		NT_Tcp_shm_closesocket(g_socket_to_add, g_socket_to_add_event);
		while (g_pProcTable[remote_iproc].sock == INVALID_SOCKET)
			Sleep(100);
	}

	ReleaseMutex(g_hAddSocketMutex);
	return 1;
}
