#include "nt_global_cpp.h"
#include <stdio.h>

HANDLE g_hControlThread, g_hControlThreadClient;
HANDLE g_hStopControlThreadEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
HANDLE g_hEveryoneConnectedEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
HANDLE g_hControlThreadClientMutex = CreateMutex(NULL, FALSE, NULL);
HANDLE g_hControlThreadClientEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
HANDLE g_hControlCommandCompletedEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
int g_nControlListenPort;
SOCKET g_hControlSocket;
WSAEVENT g_hControlSocketEvent;
char g_nControlCommand;
int g_nControlCommandData;

// Function name	: ControlThread
// Description	    : 
// Return type		: void 
// Argument         : HANDLE hReadyEvent
void ControlThread(HANDLE hReadyEvent)
{
	SOCKET sock, *aSock;
	WSAEVENT sock_event, *ahEvent;
	int error = 0, i, j;
	char host[NT_HOSTNAME_LEN];

	try{

	// create a listening socket
	error = NT_Tcp_shm_create_bind_socket(&sock, &sock_event);
	if (error)
		nt_error("ControlThread: NT_Tcp_shm_create_bind_socket failed", 1);

	// associate sock_event with sock
	WSAEventSelect(sock, sock_event, FD_ACCEPT);

	if (listen(sock, SOMAXCONN) == SOCKET_ERROR)
	{
		error = WSAGetLastError();
		nt_error("ControlThread: listen failed", error);
	}

	// get the port and local hostname for the listening socket
	error = NT_Tcp_shm_get_sock_info(sock, host, &g_nControlListenPort);
	if (error)
		nt_error("ControlThread: Unable to get host and port of listening socket", error);

	// Signal that the control port is valid
	SetEvent(hReadyEvent);

	aSock = new SOCKET[g_nNproc];
	ahEvent = new WSAEVENT[g_nNproc+1];
	ahEvent[g_nNproc] = g_hStopControlThreadEvent;
	HANDLE hTwoEvents[2];
	hTwoEvents[0] = sock_event;
	hTwoEvents[1] = g_hStopControlThreadEvent;

	// Accept connections from everyone
	for (i=0; i<g_nNproc; i++)
	{
		SOCKET temp_socket;
		DWORD result;
		while (true)
		{
			result = WSAWaitForMultipleEvents(2, hTwoEvents, FALSE, 30000, FALSE);
			if (result == WSA_WAIT_TIMEOUT)
			{
				nt_error("ControlThread: Wait for remote process to connect to process zero timed out.", 1);
				return;
			}
			if (result - WSA_WAIT_EVENT_0 == 1)
			{
				//DPRINTF(("g_hStopControlThreadEvent signalled.\n"));
				while (--i > 0)
				{
					closesocket(aSock[i]);
					WSACloseEvent(ahEvent[i]);
				}
				delete aSock;
				delete ahEvent;
				return;
			}
			temp_socket = accept(sock, NULL, NULL);
			if (temp_socket != INVALID_SOCKET)
				break;
			int error = GetLastError();
			if (error == WSAEWOULDBLOCK)
			{
				WSAResetEvent(sock_event);
				WSAEventSelect(sock, sock_event, FD_ACCEPT);
			}
			else
			{
				nt_error("ControlThread: accept failed", error);
				return;
			}
		}

		int remote_iproc=0, remote_port=0, ret_val=0;
		char remote_host[NT_HOSTNAME_LEN]="", ack=1;
		WSAEVENT temp_event = WSACreateEvent();
		if (WSAEventSelect(temp_socket, temp_event, FD_READ | FD_CLOSE) == SOCKET_ERROR)
		{
			nt_error("ControlThread: WSAEventSelect failed", WSAGetLastError());
		}
		// Receive iproc, port, hostname, exename, and pid
		if (ret_val = ReceiveBlocking(temp_socket, temp_event, (char*)&remote_iproc, sizeof(int), 0))
			nt_error("ControlThread: recv remote_iproc failed.", ret_val);
		if (ret_val = ReceiveBlocking(temp_socket, temp_event, (char*)&remote_port, sizeof(int), 0))
			nt_error("ControlThread: recv remote_port failed.", ret_val);
		if (ret_val = ReceiveBlocking(temp_socket, temp_event, remote_host, NT_HOSTNAME_LEN, 0))
			nt_error("ControlThread: recv remote_host failed.", ret_val);
		if (ret_val = ReceiveBlocking(temp_socket, temp_event, g_pProcTable[remote_iproc].exename, NT_EXENAME_LEN, 0))
			nt_error("ControlThread: recv remote_exename failed.", ret_val);
		if (ret_val = ReceiveBlocking(temp_socket, temp_event, (char*)&g_pProcTable[remote_iproc].pid, sizeof(int), 0))
			nt_error("ControlThread: recv remote_pid failed.", ret_val);
		// Send acknowledgement
		if (SendBlocking(temp_socket, &ack, 1, 0) == SOCKET_ERROR)
			nt_error("ControlThread: send ack failed.", WSAGetLastError());

		// Insert the information in aSock and ahEvent
		aSock[remote_iproc] = temp_socket;
		ahEvent[remote_iproc] = temp_event;

		// Insert hostname and listen port into g_pProcTable and signal that the data is valid
		g_pProcTable[remote_iproc].listen_port = remote_port;
		strcpy(g_pProcTable[remote_iproc].host, remote_host); 
	}

	SetEvent(g_hEveryoneConnectedEvent);

	// Loop and handle commands that arrive on the control sockets
	DWORD result;
	WSANETWORKEVENTS nevents;
	char command;
	int query_n, ret_val, num_left = g_nNproc;

	while (true)
	{
		result = WSAWaitForMultipleEvents(g_nNproc+1, ahEvent, FALSE, INFINITE, FALSE);
		for (i=0; i<g_nNproc; i++)
		{
			WSAEnumNetworkEvents(
				aSock[i],
				ahEvent[i],
				&nevents);
			if (nevents.lNetworkEvents & FD_READ)
			{
				recv(aSock[i], &command, 1, 0);
				switch (command)
				{
				case NT_TCP_CONNECT_INFO_CMD:
					//DPRINTF(("NT_TCP_CONNECT_INFO_CMD received from %d\n", i));
					if (ret_val = ReceiveBlocking(aSock[i], ahEvent[i], (char*)&query_n, sizeof(int), 0))
						nt_error("ControlThread: ReceiveBlocking query_n failed", ret_val);
					if (SendBlocking(aSock[i], g_pProcTable[query_n].host, NT_HOSTNAME_LEN, 0) == SOCKET_ERROR)
						MakeErrMsg(WSAGetLastError(), "ControlThread: send host %d failed", query_n);
					if (SendBlocking(aSock[i], (char*)&g_pProcTable[query_n].listen_port, sizeof(int), 0) == SOCKET_ERROR)
						MakeErrMsg(WSAGetLastError(), "ControlThread: send listen_port %d failed", g_pProcTable[query_n].listen_port);
					break;
				case NT_TCP_PROCESS_INFO_CMD:
					//DPRINTF(("NT_TCP_PROCESS_INFO_CMD received from %d\n", i));
					if (ret_val = ReceiveBlocking(aSock[i], ahEvent[i], (char*)&query_n, sizeof(int), 0))
						nt_error("ControlThread: ReceiveBlocking query_n failed", ret_val);
					if (SendBlocking(aSock[i], g_pProcTable[query_n].host, NT_HOSTNAME_LEN, 0) == SOCKET_ERROR)
						MakeErrMsg(WSAGetLastError(), "ControlThread: send host %d failed", query_n);
					if (SendBlocking(aSock[i], g_pProcTable[query_n].exename, NT_EXENAME_LEN, 0) == SOCKET_ERROR)
						MakeErrMsg(WSAGetLastError(), "ControlThread: send exename %d failed", query_n);
					if (SendBlocking(aSock[i], (char*)&g_pProcTable[query_n].pid, sizeof(int), 0) == SOCKET_ERROR)
						MakeErrMsg(WSAGetLastError(), "ControlThread: send process %d id %d failed", query_n, g_pProcTable[query_n].pid);
					break;
				case NT_TCP_END_CMD:
					num_left--;
					if (num_left == 0)
					{
						command = 1;
						for (j=0; j<g_nNproc; j++)
						{
							if (SendBlocking(aSock[j], &command, 1, 0) == SOCKET_ERROR)
								DPRINTF(("Unable to send char to process %d\n", j));
							NT_Tcp_shm_closesocket(aSock[j], ahEvent[j]);
						}
						delete aSock;
						delete ahEvent;
						return;
					}
					break;
				default:
					nt_error("ControlThread: Invalid command received", i);
				}
			}
			if (nevents.lNetworkEvents & FD_CLOSE)
			{
				// How should I handle a closed connection?
				command = NT_TCP_END_CMD;
				for (j=1; j<g_nNproc; j++) // skip process 0 because this is process 0
				{
					if (SendBlocking(aSock[j], &command, 1, 0) == SOCKET_ERROR)
						DPRINTF(("Unable to send char to process %d\n", j));
					NT_Tcp_shm_closesocket(aSock[j], ahEvent[j]);
				}
				delete aSock;
				delete ahEvent;
				MakeErrMsg(1, "Control socket %d closed", i);
				return;
			}
		}
	}
	}catch(int e)
	{
		nt_error("Integer exception caught in ControlThread", e);
	}
	catch(...)
	{
		nt_error("Exception caught in ControlThread.", 1);
	}
}

// Function name	: ControlThreadClient
// Description	    : 
// Return type		: void 
void ControlThreadClient()
{
	int ret_val;
	DWORD result;
	WSANETWORKEVENTS nevents;
	char command;
	HANDLE ahEvent[2];

	ahEvent[0] = g_hControlThreadClientEvent; // signalled by other threads issue commands
	ahEvent[1] = g_hControlSocketEvent;

	// Loop and handle commands to and from the control socket
	while (true)
	{
		// This wait is set to INFINITE.  Should there be a timeout?
		result = WSAWaitForMultipleEvents(2, ahEvent, FALSE, INFINITE, FALSE);

		// Check to see if the wait command failed
		if ((result < WAIT_OBJECT_0) || (result>WAIT_OBJECT_0+1))
			nt_error("Wait failed in ControlThreadClient", 1);

		// Check to see if the socket was signalled
		WSAEnumNetworkEvents(
			g_hControlSocket,
			g_hControlSocketEvent,
			&nevents);
		if (nevents.lNetworkEvents & FD_READ)
		{
			recv(g_hControlSocket, &command, 1, 0);
			switch(command)
			{
			case NT_TCP_END_CMD:
				nt_error("end command received over control thread", 1);
				break;
			default:
				nt_error("unexpected message sent over control socket", 1);
				break;
			}
		}
		if (nevents.lNetworkEvents & FD_CLOSE)
		{
			nt_error("Control socket closed", 1);
		}

		// Check to see if the inter thread event was signalled
		if (WaitForSingleObject(g_hControlThreadClientEvent, 0) == WAIT_OBJECT_0)
		{
			switch(g_nControlCommand)
			{
			case NT_TCP_CONNECT_INFO_CMD:
				// Ask the root process how to reach process 'g_nControlCommandData'
				if (SendBlocking(g_hControlSocket, &g_nControlCommand, 1, 0) == SOCKET_ERROR)
					nt_error("send NT_TCP_CONNECT_INFO_CMD failed in ControlThreadClient", WSAGetLastError());
				if (SendBlocking(g_hControlSocket, (char*)&g_nControlCommandData, sizeof(int), 0) == SOCKET_ERROR)
					nt_error("send remote_iproc failed in ControlThreadClient", WSAGetLastError());
				if (ret_val = ReceiveBlocking(g_hControlSocket, g_hControlSocketEvent, g_pProcTable[g_nControlCommandData].host, NT_HOSTNAME_LEN, 0))
					MakeErrMsg(ret_val, "ReceiveBlocking host(%d) failed in ConnectTo", g_nControlCommandData);
				if (ret_val = ReceiveBlocking(g_hControlSocket, g_hControlSocketEvent, (char*)&g_pProcTable[g_nControlCommandData].listen_port, sizeof(int), 0))
					nt_error("ReceiveBlocking listen_port failed in ConnectTo", ret_val);
				
				ResetEvent(g_hControlThreadClientEvent);
				SetEvent(g_hControlCommandCompletedEvent);
				break;
			case NT_TCP_PROCESS_INFO_CMD:
				// Send request to root for information on process 'g_nControlCommandData'
				if (SendBlocking(g_hControlSocket, &g_nControlCommand, 1, 0) == SOCKET_ERROR)
					MakeErrMsg(1, "send NT_TCP_PROCESS_INFO_CMD command to root process failed, querying process %d", g_nControlCommandData);
				if (SendBlocking(g_hControlSocket, (char*)&g_nControlCommandData, sizeof(int), 0) == SOCKET_ERROR)
					MakeErrMsg(1, "send integer %d after NT_TCP_PROCESS_INFO_CMD failed", g_nControlCommandData);
				
				// Receive information
				ReceiveBlocking(g_hControlSocket, g_hControlSocketEvent, g_pProcTable[g_nControlCommandData].host, NT_HOSTNAME_LEN, 0);
				ReceiveBlocking(g_hControlSocket, g_hControlSocketEvent, g_pProcTable[g_nControlCommandData].exename, NT_EXENAME_LEN, 0);
				ReceiveBlocking(g_hControlSocket, g_hControlSocketEvent, (char*)&g_pProcTable[g_nControlCommandData].pid, sizeof(int), 0);

				ResetEvent(g_hControlThreadClientEvent);
				SetEvent(g_hControlCommandCompletedEvent);
				break;
			case NT_TCP_END_CMD:
				if (SendBlocking(g_hControlSocket, &g_nControlCommand, 1, 0) == SOCKET_ERROR)
					nt_error("send command over g_hControlSocket failed in ControlThreadClient", WSAGetLastError());
				if (ReceiveBlocking(g_hControlSocket, g_hControlSocketEvent, &g_nControlCommand, 1, 0))
					nt_error("ReceiveBlocking from g_hControlSocket failed in ControlThreadClient", 1);
				NT_Tcp_shm_closesocket(g_hControlSocket, g_hControlSocketEvent);

				ResetEvent(g_hControlThreadClientEvent);
				SetEvent(g_hControlCommandCompletedEvent);
				return;
				break;
			default:
				MakeErrMsg(1, "g_nControlCommand invalid: %d", g_nControlCommand);
				break;
			}
		}
	}
}

// Function name	: ConnectToRoot
// Description	    : 
// Return type		: void 
// Argument         : char *host
// Argument         : int port
void ConnectToRoot(char *host, int port)
{
	int ret_val;
	char ack;

	// create the event
	g_hControlSocketEvent = WSACreateEvent();
	if (g_hControlSocketEvent == WSA_INVALID_EVENT)
		nt_error("WSACreateEvent failed in ConnectToRoot", WSAGetLastError());
	// create the socket
	g_hControlSocket = socket(PF_INET, SOCK_STREAM, 0);
	if (g_hControlSocket == INVALID_SOCKET)
		nt_error("socket failed in ConnectToRoot", WSAGetLastError());

	//printf("Connecting to root: %s %d\n", host, port);fflush(stdout);
	ret_val = NT_Tcp_shm_connect(g_hControlSocket, host, port);
	if (ret_val)
		nt_error("ConnectToRoot: NT_Tcp_shm_connect failed", ret_val);

	if (WSAEventSelect(g_hControlSocket, g_hControlSocketEvent, FD_READ | FD_CLOSE) == SOCKET_ERROR)
		nt_error("ConnectToRoot: WSAEventSelect failed", WSAGetLastError());

	// Send iproc, listen_port, hostname, exename, and pid
	if (SendBlocking(g_hControlSocket, (char*)&g_nIproc, sizeof(int), 0) == SOCKET_ERROR)
		nt_error("ConnectToRoot: send iproc failed", WSAGetLastError());
	if (SendBlocking(g_hControlSocket, (char*)&g_pProcTable[g_nIproc].listen_port, sizeof(int), 0) == SOCKET_ERROR)
		nt_error("ConnectToRoot: send listen port failed", WSAGetLastError());
	if (SendBlocking(g_hControlSocket, g_pszHostName, NT_HOSTNAME_LEN, 0) == SOCKET_ERROR)
		nt_error("ConnectToRoot: send host name failed", WSAGetLastError());
	if (SendBlocking(g_hControlSocket, g_pProcTable[g_nIproc].exename, NT_EXENAME_LEN, 0) == SOCKET_ERROR)
		nt_error("ConnectToRoot: send exe name failed", WSAGetLastError());
	if (SendBlocking(g_hControlSocket, (char*)&g_pProcTable[g_nIproc].pid, sizeof(int), 0) == SOCKET_ERROR)
		nt_error("ConnectToRoot: send pid failed", WSAGetLastError());

	// Wait for an ack to ensure the data was received
	ret_val = ReceiveBlocking(g_hControlSocket, g_hControlSocketEvent, &ack, 1, 0);
	if (ret_val)
		nt_error("ConnectToRoot: recv ack failed", WSAGetLastError());

	// Start up the client thread to monitor control messages and events
	DWORD dwThreadID;
	g_hControlThreadClient = CreateThread(
		NULL, 0, (LPTHREAD_START_ROUTINE)ControlThreadClient, NULL, 0, &dwThreadID);
	if (g_hControlThreadClient == NULL)
		nt_error("CreateThread for the ControlThreadClient failed", GetLastError());
}
