#include "nt_global_cpp.h"
#include <stdio.h>

HANDLE g_hListenThread;
HANDLE g_hStopListenThreadEvent = CreateEvent(NULL, TRUE, FALSE, NULL);

void ListenThread(HANDLE hReadyEvent)
{
	SOCKET sock;
	WSAEVENT ahEvent[2];
	int error = 0;

	ahEvent[1] = g_hStopListenThreadEvent;

	// create a listening socket
	error = NT_Tcp_shm_create_socket(&sock, &ahEvent[0]);
	if (error)
		nt_error("ListenThread: NT_Tcp_shm_create_socket failed", 1);

	// associate listen_socket_event with listen_socket
	WSAEventSelect(sock, ahEvent[0], FD_ACCEPT);

	if (listen(sock, SOMAXCONN) == SOCKET_ERROR)
		nt_error("ListenThread: listen failed", WSAGetLastError());

	// get the port and local hostname for the listening socket
	error = NT_Tcp_shm_get_sock_info(sock, g_pProcTable[g_nIproc].host, &g_pProcTable[g_nIproc].listen_port);
	if (error)
		nt_error("ListenThread: Unable to get host and port of listening socket", error);

	// Signal that the port number is valid
	SetEvent(hReadyEvent);

	try{
	while (true)
	{
		SOCKET temp_socket;
		WSAEVENT temp_event;
		DWORD ret_val;
		int error, remote_iproc;

		// Wait and accept a remote connection
		// This wait is set for INFINITE.  If a timeout were desired, how long
		// would be appropriate to wait for the launched processes to contact 
		// the root - seconds, minutes?
		ret_val = WSAWaitForMultipleEvents(2, ahEvent, FALSE, INFINITE, FALSE);
		if (ret_val == WSA_WAIT_FAILED)
		{
			nt_error("ListenThread: Wait for remote process to connect failed", ret_val);
			return;
		}
		else
		if (ret_val == WSA_WAIT_EVENT_0)
		{
			temp_socket = accept(sock, NULL, NULL);
			if (temp_socket != INVALID_SOCKET)
			{
				temp_event = WSACreateEvent();
				WSAEventSelect(temp_socket, temp_event, FD_READ | FD_CLOSE);
				
				// Receive iproc
				ReceiveBlocking(temp_socket, temp_event, (char*)&remote_iproc, sizeof(int), 0);
				
				// Insert the information in g_pProcTable
				g_pProcTable[remote_iproc].sock_event = temp_event;
				g_pProcTable[remote_iproc].sock = temp_socket;
				//DPRINTF(("process %d: socket inserted in location %d\n", g_nIproc, remote_iproc));
				//LogMsg("process %d: Connection accepted from %d\n", g_nIproc, remote_iproc);
				break;
			}
			error = WSAGetLastError();
			if (error == WSAEWOULDBLOCK)
			{
				//fdebug("WSAEWOULDBLOCK while trying to accept connections\n");
				WSAResetEvent(ahEvent[0]);
				WSAEventSelect(sock, ahEvent[0], FD_ACCEPT);
			}
			else
			{
				nt_error("ListenThread: accept failed", error);
				return;
			}
		}
		else
		if (ret_val - WSA_WAIT_EVENT_0 == 1)
		{
			closesocket(sock);
			WSACloseEvent(ahEvent[0]);
			DPRINTF(("ListenThread signalled to stop.\n"));
			return;
		}
		else
		{
			nt_error("ListenThread: Unknown error waiting for connections", ret_val);
		}
	}
	}catch(...)
	{
		nt_error("Exception caught in ListenThread.", 1);
	}
}
