// nt_tcp_shm_message_queue.h: interface for the MessageQueue class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_MESSAGEQUEUE_H__BF8A68B9_19C3_11D3_95D2_009027106653__INCLUDED_)
#define AFX_MESSAGEQUEUE_H__BF8A68B9_19C3_11D3_95D2_009027106653__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include <windows.h>

#define MSGQ_ELEMENT_BLOCK 1000

class MessageQueue  
{
public:
	struct MsgQueueElement
	{
		HANDLE hEvent;
		bool bIn_use;
	};

	MessageQueue();
	virtual ~MessageQueue();

	void* GetBuffer(int tag, int length, int from, MsgQueueElement **ppElement);
	bool FillBuffer(int tag, void *buffer, int *length, int *from);
	bool Available(int tag, int &from);

	bool PostRecvBuffer(int tag, void *buffer, int length, int *pID);
	bool Wait(int *pID);
	bool Test(int *pID);
private:
	struct InternalListNode
	{
		int length, from;
		void *buffer;
		MsgQueueElement *element;
		InternalListNode *next;
	};
	struct InternalNode
	{
		int tag;
		InternalListNode *list, *posted;
		InternalNode *left;
		InternalNode *right;
	};

	// Buffer management variables
	CRITICAL_SECTION m_CriticalSection;
	InternalNode *m_pHeap;

	// Event management variables
	int m_num_elements;
	int m_num_available;
	int m_cur_index;
	int m_pool_size;
	MsgQueueElement **m_pPool;

	// Event management functions
	MsgQueueElement* allocElement();
	void freeElement(MsgQueueElement *element);

	// global variables used for FindNode
	int find_tag, find_length, find_from;
	void *find_buffer;
	MsgQueueElement *find_pElement;

	void FindNode(InternalNode *&node);
	bool FindAvailable(InternalNode *node, int tag, int &from);

	InternalListNode *fillfind_node;
	void FillFindNode(InternalNode *&node);

	// Cleanup functions
	void DeleteHeapList(InternalListNode *listnode);
	void DeleteHeap(InternalNode *node);
};

#endif // !defined(AFX_MESSAGEQUEUE_H__BF8A68B9_19C3_11D3_95D2_009027106653__INCLUDED_)
